/*
 * Name:    mt_record.c
 *
 * Purpose: Tcl interface to generic MX record functions.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2002 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include <tcl.h>
#include <itcl.h>

#include "mx_util.h"
#include "mx_version.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_handle.h"
#include "mx_log.h"
#include "mx_tcl.h"
#include "mx_tcl_private.h"

MX_EXPORT mx_status_type
mxtcl_get_record_from_object( MX_RECORD **record,
				Tcl_Interp *interp,
				Tcl_Obj *CONST obj )
{
	const char fname[] = "mxtcl_get_record_from_object()";

	signed long record_handle;
	long record_handle_long;
	void *pointer;
	int tcl_status;
	mx_status_type mx_status;

	if ( obj == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL object pointer passed." );
	}
	if ( record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"NULL pointer to MX_RECORD pointer passed." );
	}

	*record = NULL;

	/* Get the handle for the record we were passed. */

	tcl_status = Tcl_GetLongFromObj(interp, obj, &record_handle_long);

	if ( tcl_status != TCL_OK ) {
		if ( obj->bytes == NULL ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Cannot interpret the object passed as a long.  "
		"The string representation of the object is not available.");
		} else {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Cannot interpret the object '%s' as a long.",
			obj->bytes );
		}
	}

	record_handle = (signed long) record_handle_long;

	/* Convert it to an MX_RECORD pointer. */

	mx_status = mx_get_pointer_from_handle( &pointer,
					mxtcl_global_record_handle_table,
					record_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	*record = (MX_RECORD *) pointer;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT int
mxtcl_error( Tcl_Interp *interp, mx_status_type mx_status )
{
	Tcl_Obj *resultPtr;

	resultPtr = Tcl_GetObjResult( interp );
	Tcl_SetStringObj( resultPtr, mx_status.message, -1 );

	return TCL_ERROR;
}

MX_EXPORT int
mxtcl_object_parse_error( Tcl_Interp *interp,
				Tcl_Obj *CONST obj,
				const char *type_name,
				const char *calling_fname )
{
	mx_status_type mx_status;

	if ( obj->bytes == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, calling_fname,
			"Cannot interpret the object passed as a %s.  "
			"The string representation of the object "
			"is not available.", type_name );
	} else {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, calling_fname,
			"Cannot interpret the object '%s' as a %s.",
			obj->bytes, type_name );
	}

	return mxtcl_error( interp, mx_status );
}

/*========*/

MX_EXPORT int
Mx_create_record_list_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_create_record_list_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record_list;
	MX_LIST_HEAD *list_head_struct;
	signed long record_list_handle;
	mx_status_type mx_status;

	/* Create the record list.  The pointer 'record_list' points
	 * to the first record in the list.
	 */

	record_list = mx_initialize_record_list();

	if ( record_list == NULL ) {
		(void) mx_error( MXE_FUNCTION_FAILED, fname,
		"Unable to create a new record list." );

		return TCL_ERROR;
	}

	MX_DEBUG( 2,("mx_initialize_record_list() succeeded."));

	/* If the list head structure is not set up correctly, we cannot
	 * continue.
	 */

	list_head_struct = mx_get_record_list_head_struct( record_list );

	if ( list_head_struct == (MX_LIST_HEAD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"List head structure for record list %p is NULL.",
			record_list );
	}

	/* We can't pass a C pointer back to Tcl, so we must convert
	 * the C pointer 'record_list' into an MX record handle which
	 * is an integer number.
	 */

	/* Create a handle for the record list as a whole. */

	mx_status = mx_create_handle( &record_list_handle,
					mxtcl_global_record_handle_table,
					record_list );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	record_list->handle = record_list_handle;

	list_head_struct->handle_table = mxtcl_global_record_handle_table;

	MX_DEBUG( 2,("%s: record_list = %p, record_list_handle = %ld",
			fname, record_list, record_list_handle));

	/* Now pass the record list handle back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );

	Tcl_SetLongObj( resultPtr, (long) record_list_handle );

	return TCL_OK;
}

MX_EXPORT int
Mx_read_database_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_read_database_cmd()";

	MX_RECORD *record, *record_list;
	char *database_name;
	mx_status_type mx_status;

	if ( objc != 3 ) {
		Tcl_WrongNumArgs( interp, 2, objv,
    "Usage: { record_list } read_database { database_name }" );
		return TCL_ERROR;
	}

	/* Get the record pointer for the record handle we were passed. */

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Make sure we are using the list head record. */

	record_list = record->list_head;

	/* Get the name of the MX database to be read. */

	database_name = Tcl_GetStringFromObj( objv[2], NULL );

	if ( database_name == NULL ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Second argument to MxRecordList::read_database was not a string." );
		return TCL_ERROR;
	}

	MX_DEBUG(-2,("Reading database '%s'", database_name));

	mx_status = mx_read_database_file( record_list, database_name, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	MX_DEBUG(-2,("Database successfully read."));

	return TCL_OK;
}

MX_EXPORT int
Mx_finish_database_initialization_cmd(
				ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	MX_RECORD *record, *record_list;
	mx_status_type mx_status;

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
		"Usage: { record_list } finish_database_initialization" );
		return TCL_ERROR;
	}

	/* Get the record pointer for the record handle we were passed. */

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Make sure we are using the list head record. */

	record_list = record->list_head;

	mx_status = mx_finish_database_initialization( record_list );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_initialize_hardware_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_initialize_hardware_cmd()";

	MX_RECORD *record, *record_list;
	int inithw_flags, status;
	mx_status_type mx_status;

	if ( objc != 3 ) {
		Tcl_WrongNumArgs( interp, 2, objv,
		"Usage: { record_list } initialize_hardware inithw_flags" );
		return TCL_ERROR;
	}

	/* Get the record pointer for the record handle we were passed. */

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Make sure we are using the list head record. */

	record_list = record->list_head;

	/* Get the initialize hardware flags. */

	status = Tcl_GetIntFromObj( interp, objv[2], &inithw_flags );

	if ( status != TCL_OK ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Second argument was not an integer." );

		return mxtcl_error( interp, mx_status );
	}

	mx_status = mx_initialize_hardware( record_list, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_create_record_from_description_cmd(
				ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_create_record_from_description_cmd()";

	MX_RECORD *record, *record_list, *new_record;
	char *description;
	mx_status_type mx_status;

	if ( objc != 3 ) {
		Tcl_WrongNumArgs( interp, 2, objv,
	"Usage: { record_list } create_record_from_description description" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Make sure we are using the list head record. */

	record_list = record->list_head;

	/* Get the string description. */

	description = Tcl_GetStringFromObj( objv[2], NULL );

	if ( description == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Second argument was not a string.  "
		"It was supposed to be a record description" );

		return mxtcl_error( interp, mx_status );
	}

	mx_status = mx_create_record_from_description( record_list,
						description, &new_record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_finish_record_initialization_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	MX_RECORD *record;
	mx_status_type mx_status;

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
		"Usage: { record_list } finish_record_initialization" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_delete_record_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	MX_RECORD *record;
	mx_status_type mx_status;

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
		"Usage: { record_list } delete_record" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_delete_record( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_record_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_record_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record, *record_list;
	char *record_name;
	signed long handle;
	int tcl_status, quiet_flag;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 4 ) {
		Tcl_WrongNumArgs( interp, 3, objv,
	    "Usage: { record_list } get_record { record_name quiet_flag }" );
		return TCL_ERROR;
	}

	/* Get the MX_RECORD pointer for the record list. */

	mx_status = mxtcl_get_record_from_object(
					&record_list, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Get the requested record name. */

	record_name = Tcl_GetStringFromObj( objv[2], NULL );

	if ( record_name == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Second argument was not a string.  It was supposed to be the record name" );

		return mxtcl_error( interp, mx_status );
	}

	/* Get the quiet flag. */

	tcl_status = Tcl_GetIntFromObj( interp, objv[3], &quiet_flag );

	if ( tcl_status != TCL_OK )
		return mxtcl_object_parse_error(
					interp, objv[3], "integer", fname );

	/* See if we can find this record in the specified record list. */

	record = mx_get_record( record_list, record_name );

	if ( record == NULL ) {
		if ( quiet_flag ) {
			mx_status = mx_error_quiet( MXE_NOT_FOUND, fname,
			"The record '%s' was not found in the database.",
				record_name );
		} else {
			mx_status = mx_error( MXE_NOT_FOUND, fname,
			"The record '%s' was not found in the database.",
				record_name );
		}

		return mxtcl_error( interp, mx_status );
	}

	/* Convert the record pointer to a handle. */

	mx_status = mx_get_handle_from_pointer( &handle,
				mxtcl_global_record_handle_table, record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Pass the record handle back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );

	Tcl_SetLongObj( resultPtr, (long) handle );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_next_record_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_next_record_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record, *next_record;
	signed long next_handle;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: { record } get_next_record" );
		return TCL_ERROR;
	}

	/* Get the next record in the record list. */

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	next_record = record->next_record;

	/* Convert the record pointer to a handle. */

	mx_status = mx_get_handle_from_pointer( &next_handle,
			mxtcl_global_record_handle_table, next_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Pass the record handle back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );

	Tcl_SetLongObj( resultPtr, (long) next_handle );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_record_name_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_record_name_cmd()";

	MX_RECORD *record;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: get_record_name record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	sprintf( interp->result, "%s", record->name );

	return TCL_OK;
}

MX_EXPORT int
Mx_create_description_from_record_cmd(
				ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_create_description_from_record_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record;
	char buffer[1024];
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: create_description_from_record record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_create_description_from_record(
				record, buffer, sizeof(buffer) );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Send the string back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );
	Tcl_SetStringObj( resultPtr, buffer, -1 );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_field_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_field_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	char token_buffer[1024];
	char *field_name, *ptr;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 3 ) {
		Tcl_WrongNumArgs( interp, 2, objv,
	    "Usage: get_record_name record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	field_name = Tcl_GetStringFromObj( objv[2], NULL );

	if ( field_name == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Second argument to Mx::Record::get_field was not a string." );

		return mxtcl_error( interp, mx_status );
	}

	if ( record->num_record_fields <= 0 ) {
		resultPtr = Tcl_GetObjResult( interp );
		Tcl_SetStringObj( resultPtr, "Unsupported", -1 );

		return TCL_OK;
	}

	mx_status = mx_find_record_field( record, field_name, &record_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_create_description_from_field( record, record_field,
			token_buffer, sizeof( token_buffer ) - 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* There may be an extra space in front if we invoked
	 * mx_create_array_description().
	 */

	ptr = &token_buffer[0];

	if ( *ptr == ' ' ) {
		ptr++;
	}

	/* Send the string back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );
	Tcl_SetStringObj( resultPtr, ptr, -1 );

	return TCL_OK;
}

MX_EXPORT int
Mx_set_field_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_set_field_cmd()";

	mx_status_type ( *token_parser ) ( void *, char *,
				MX_RECORD *, MX_RECORD_FIELD *,
				MX_RECORD_FIELD_PARSE_STATUS * );
	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD_PARSE_STATUS temp_parse_status;
	char *token_buffer, *field_name;
	void *value_ptr;
	mx_status_type mx_status;

	static char separators[] = MX_RECORD_FIELD_SEPARATORS;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 4 ) {
		Tcl_WrongNumArgs( interp, 3, objv,
	    "Usage: get_record_name record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	field_name = Tcl_GetStringFromObj( objv[2], NULL );

	if ( field_name == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Second argument to Mx::Record::set_field was not a string." );

		return mxtcl_error( interp, mx_status );
	}

	MX_DEBUG( 2,("%s: field_name = '%s'", fname, field_name));

	if ( record->num_record_fields <= 0 ) {
		sprintf( interp->result, "Unsupported" );

		return TCL_OK;
	}

	mx_status = mx_find_record_field( record, field_name, &record_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	value_ptr = mx_get_field_value_pointer( record_field );

	mx_status = mx_get_token_parser( record_field->datatype,
						&token_parser );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* Parse the string description of this field. */

	token_buffer = Tcl_GetStringFromObj( objv[3], NULL );

	if ( token_buffer == NULL ) {
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Third argument to Mx::Record::set_field was not a string." );

		return mxtcl_error( interp, mx_status );
	}

	MX_DEBUG( 2,("%s: token_buffer = '%s'", fname, token_buffer));

	temp_parse_status.description = token_buffer;
	temp_parse_status.separators = separators;
	temp_parse_status.num_separators = strlen(separators);
	temp_parse_status.start_of_trailing_whitespace = NULL;

	if ( record_field->datatype == MXFT_STRING ) {
		temp_parse_status.max_string_token_length =
				mx_get_max_string_token_length( record_field );
	} else {
		temp_parse_status.max_string_token_length = 0L;
	}

	if ( (record_field->num_dimensions == 0)
	  || ((record_field->datatype == MXFT_STRING)
	     && (record_field->num_dimensions == 1)) ) {

		mx_status = mx_get_next_record_token( &temp_parse_status,
				token_buffer, 1 + strlen(token_buffer) );

		MX_DEBUG( 2,("%s: single token = '%s'", fname, token_buffer));

		mx_status = (*token_parser)( value_ptr,
				token_buffer, NULL, record_field,
				&temp_parse_status );
	} else {
		mx_status = mx_parse_array_description( value_ptr,
				record_field->num_dimensions - 1, NULL,
				record_field, &temp_parse_status, token_parser);
	}

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_object_class_name_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_object_class_name_cmd()";

	Tcl_Obj *resultPtr;
	MX_RECORD *record;
	char buffer[80];
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: get_object_class_name record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	/* For an unknown type of record, the object name is Mx::Record. */

	strcpy( buffer, "Mx::Record" );

	/* For many types of records, we override this. */

	switch( record->mx_superclass ) {
	case MXR_INTERFACE:
		switch( record->mx_class ) {
		case MXI_RS232:
			strcpy( buffer, "Mx::RS232" );
			break;
		}
		break;
	case MXR_DEVICE:
		switch( record->mx_class ) {
		case MXC_DIGITAL_INPUT:
			strcpy( buffer, "Mx::DigitalInput" );
			break;
		case MXC_DIGITAL_OUTPUT:
			strcpy( buffer, "Mx::DigitalOutput" );
			break;
		case MXC_MOTOR:
			strcpy( buffer, "Mx::Motor" );
			break;
		case MXC_RELAY:
			strcpy( buffer, "Mx::Relay" );
			break;
		case MXC_SCALER:
			strcpy( buffer, "Mx::Scaler" );
			break;
		case MXC_TIMER:
			strcpy( buffer, "Mx::Timer" );
			break;
		}
		break;
	case MXR_SCAN:
		strcpy( buffer, "Mx::Scan" );
		break;
	case MXR_VARIABLE:
		strcpy( buffer, "Mx::Variable" );
		break;
	}

	/* Send the string back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );
	Tcl_SetStringObj( resultPtr, buffer, -1 );

	return TCL_OK;
}

MX_EXPORT int
Mx_get_version_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_get_version_cmd()";

	Tcl_Obj *resultPtr;
	int major, minor, update;
	char *date;
	char buffer[80];

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 1 ) {
		Tcl_WrongNumArgs( interp, 0, objv, "Usage: get_version" );
		return TCL_ERROR;
	}

	major  = mx_get_major_version();
	minor  = mx_get_minor_version();
	update = mx_get_update_version();
	date   = mx_get_version_date();

	sprintf( buffer, "%d.%d.%d (%s)", major, minor, update, date );

	/* Send the string back to Tcl. */

	resultPtr = Tcl_GetObjResult( interp );
	Tcl_SetStringObj( resultPtr, buffer, -1 );

	return TCL_OK;
}

MX_EXPORT int
Mx_log_open_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_log_open_cmd()";

	MX_RECORD *record;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: log_open record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_log_open( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

MX_EXPORT int
Mx_log_close_cmd( ClientData clientData, Tcl_Interp *interp,
				int objc, Tcl_Obj *CONST objv[] )
{
	const char fname[] = "Mx_log_close_cmd()";

	MX_RECORD *record;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( objc != 2 ) {
		Tcl_WrongNumArgs( interp, 1, objv,
	    "Usage: log_close record_handle" );
		return TCL_ERROR;
	}

	mx_status = mxtcl_get_record_from_object( &record, interp, objv[1] );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	mx_status = mx_log_close( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mxtcl_error( interp, mx_status );

	return TCL_OK;
}

